from __future__ import annotations

from typing import Container, Optional

from sqlalchemy.engine import Connection

from sqlalchemy_declarative_extensions.dialects.postgresql.acl import (
    parse_acl,
    parse_default_acl,
)
from sqlalchemy_declarative_extensions.dialects.postgresql.role import Role
from sqlalchemy_declarative_extensions.dialects.postgresql.schema import (
    default_acl_query,
    object_acl_query,
    objects_query,
    roles_query,
    schema_exists_query,
    schemas_query,
)


def get_schemas_postgresql(connection: Connection):
    from sqlalchemy_declarative_extensions.schema.base import Schema

    return {
        Schema(schema) for schema, *_ in connection.execute(schemas_query).fetchall()
    }


def check_schema_exists_postgresql(connection: Connection, name: str) -> bool:
    row = connection.execute(schema_exists_query, schema=name).scalar()
    return not bool(row)


def get_objects_postgresql(connection: Connection):
    return sorted(
        [
            (r.schema, qualify_name(r.schema, r.object_name), r.relkind)
            for r in connection.execute(objects_query).fetchall()
        ]
    )


def get_default_grants_postgresql(
    connection: Connection,
    roles: Optional[Container[str]] = None,
    expanded: bool = False,
):
    default_permissions = connection.execute(default_acl_query).fetchall()
    current_role: str = connection.engine.url.username  # type: ignore

    result = []
    for permission in default_permissions:
        for acl_item in permission.acl:
            default_grants = parse_default_acl(
                acl_item,
                permission.object_type,
                permission.schema_name,
                current_role=current_role,
                expanded=expanded,
            )
            for default_grant in default_grants:
                if roles is None or default_grant.grant.target_role in roles:
                    result.append(default_grant)

    return result


def get_grants_postgresql(
    connection: Connection,
    roles: Optional[Container[str]] = None,
    expanded=False,
):
    existing_permissions = connection.execute(object_acl_query).fetchall()

    result = []
    for permission in existing_permissions:
        acl = permission.acl
        if acl is None:
            acl = [acl]

        for acl_item in acl:
            grants = parse_acl(
                acl_item,
                permission.relkind,
                qualify_name(permission.schema, permission.name),
                owner=permission.owner,
                expanded=expanded,
            )
            for grant in grants:
                if roles is None or grant.grant.target_role in roles:
                    result.append(grant)

    return result


def get_roles_postgresql(connection: Connection, exclude=None):
    result = [
        Role.from_pg_role(r)
        for r in connection.execute(roles_query).fetchall()
        if not r.rolname.startswith("pg_")
    ]
    if exclude:
        result = [role for role in result if role.name not in exclude]
    return result


def qualify_name(schema, name):
    if not schema or schema == "public":
        return name
    return f"{schema}.{name}"



from sqlalchemy_declarative_extensions import dialects
from sqlalchemy_declarative_extensions.alembic import register_alembic_events
from sqlalchemy_declarative_extensions.grant import Grants
from sqlalchemy_declarative_extensions.role import Role
from sqlalchemy_declarative_extensions.role.base import Roles
from sqlalchemy_declarative_extensions.row import Row, Rows
from sqlalchemy_declarative_extensions.schema import Schema, Schemas

# isort: split
from sqlalchemy_declarative_extensions.api import (
    declarative_database,
    declare_database,
    register_sqlalchemy_events,
)

__all__ = [
    "declarative_database",
    "declare_database",
    "Grants",
    "Role",
    "Row",
    "Rows",
    "dialects",
    "register_alembic_events",
    "register_sqlalchemy_events",
    "Role",
    "Roles",
    "Schema",
    "Schemas",
]



from sqlalchemy_declarative_extensions.dialects import postgresql
from sqlalchemy_declarative_extensions.dialects.query import (
    check_schema_exists,
    get_default_grants,
    get_grants,
    get_objects,
    get_role_cls,
    get_roles,
    get_schemas,
)

__all__ = [
    "postgresql",
    "check_schema_exists",
    "get_default_grants",
    "get_grants",
    "get_roles",
    "get_objects",
    "get_role_cls",
    "get_schemas",
]



from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import Iterable, List, Optional, Union

from sqlalchemy_declarative_extensions.dialects import postgresql

G = Union[postgresql.DefaultGrantStatement, postgresql.GrantStatement]


@dataclass
class Grants:
    """A collection of grants and the settings for diff/collection.

    Arguments:
        grants: The list of grants
        ignore_unspecified: Defaults to `False`. When `True`, ignore detected grants
            which do not match the set of defined grants.
        ignore_self_grants: Defaults to `True`. When `True`, ignores grants to the
            current user. It's typical in migrations that the a single user performs
            migrations and will have implicitly granted grants on all objects. In this
            scenario, it can be tedious to define those permissions on every object,
            so they are ignored by default.
        only_defined_roles: Defaults to `True`. When `True`, only applies to roles
            specified in the `roles` section.
        default_grants_imply_grants: Defaults to `True`. When `True`, default grants
            also imply the set of expected actual grants. This allows one to specify
            only default grants, and per-object grants will be made to match the
            default set.

    Examples:
        - No grants

        >>> grants = Grants()

        - Some options set

        >>> grants = Grants(ignore_unspecified=True)

        - With some actual grants

        >>> from sqlalchemy_declarative_extensions.dialects.postgresql import DefaultGrant
        >>> grants = Grants().are(DefaultGrant(..., ...), ...)
    """

    grants: List[G] = field(default_factory=list)

    ignore_unspecified: bool = False
    ignore_self_grants: bool = True
    only_defined_roles: bool = True
    default_grants_imply_grants: bool = True

    @classmethod
    def coerce_from_unknown(
        cls, unknown: Union[None, Iterable[G], Grants]
    ) -> Optional[Grants]:
        if isinstance(unknown, Grants):
            return unknown

        if isinstance(unknown, Iterable):
            return Grants().are(*unknown)

        return None

    def __iter__(self):
        for grant in self.grants:
            yield grant

    def are(self, *grants: G):
        return replace(self, grants=list(grants))



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

from alembic.operations import Operations
from sqlalchemy.engine import Connection

from sqlalchemy_declarative_extensions.dialects import get_role_cls, get_roles
from sqlalchemy_declarative_extensions.role.base import Roles
from sqlalchemy_declarative_extensions.role.generic import Role
from sqlalchemy_declarative_extensions.role.topological_sort import topological_sort


class RoleOp:
    pass


@dataclass
class CreateRoleOp(RoleOp):
    role: Role

    @classmethod
    def create_role(cls, operations: Operations, role_name: str, **options):
        assert operations.migration_context.connection
        role_cls = get_role_cls(operations.migration_context.connection)
        role = role_cls(role_name, **options)
        op = cls(role)
        return operations.invoke(op)  # type: ignore

    def reverse(self):
        return DropRoleOp(self.role)

    def to_sql(self):
        return self.role.to_sql_create()


@dataclass
class UpdateRoleOp(RoleOp):
    from_role: Role
    to_role: Role

    @classmethod
    def update_role(
        cls, operations, role_name: str, *, from_options=None, **to_options
    ):
        assert operations.migration_context.connection
        role_cls = get_role_cls(operations.migration_context.connection)

        from_role = role_cls(role_name, **(from_options or {}))
        to_role = role_cls(role_name, **to_options)
        op = cls(from_role, to_role)
        return operations.invoke(op)

    def reverse(self):
        return UpdateRoleOp(from_role=self.to_role, to_role=self.from_role)

    def to_sql(self):
        return self.from_role.to_sql_update(self.to_role)


@dataclass
class DropRoleOp(RoleOp):
    role: Role

    @classmethod
    def drop_role(cls, operations, role_name: str):
        op = cls(Role(role_name))
        return operations.invoke(op)

    def reverse(self):
        return CreateRoleOp(self.role)

    def to_sql(self):
        return self.role.to_sql_drop()


Operation = Union[CreateRoleOp, UpdateRoleOp, DropRoleOp]


def compare_roles(connection: Connection, roles: Roles) -> List[Operation]:
    result: List[Operation] = []
    if not roles:
        return result

    roles_by_name = {r.name: r for r in roles.roles}
    expected_role_names = set(roles_by_name)

    existing_roles = get_roles(connection)
    existing_roles_by_name = {r.name: r for r in existing_roles}
    existing_role_names = set(existing_roles_by_name)

    new_role_names = expected_role_names - existing_role_names
    removed_role_names = existing_role_names - expected_role_names

    role_cls = get_role_cls(connection)

    for role in topological_sort(roles.roles):
        role_name = role.name

        if role_name in roles.ignore_roles:
            continue

        role_created = role_name in new_role_names

        if role_created:
            result.append(CreateRoleOp(role))
        else:
            existing_role = existing_roles_by_name[role_name]
            role_cls = type(existing_role)

            # An input role might be defined as a more general `Role` while
            # the `existing_role` will always be a concrete dialect-specific version.
            concrete_defined_role = role_cls.from_unknown_role(role)

            role_updated = existing_role != concrete_defined_role
            if role_updated:
                existing_role = existing_roles_by_name[role_name]
                result.append(
                    UpdateRoleOp(
                        from_role=existing_role,
                        to_role=role_cls.from_unknown_role(role),
                    )
                )

    if not roles.ignore_unspecified:
        for removed_role in removed_role_names:
            if removed_role in roles.ignore_roles:
                continue
            result.append(DropRoleOp(role_cls(removed_role)))

    return result



from alembic.autogenerate.compare import comparators
from alembic.autogenerate.render import renderers
from alembic.operations import Operations
from sqlalchemy.schema import CreateSchema, DropSchema

from sqlalchemy_declarative_extensions import schema
from sqlalchemy_declarative_extensions.schema.base import Schemas
from sqlalchemy_declarative_extensions.schema.compare import (
    CreateSchemaOp,
    DropSchemaOp,
)

Operations.register_operation("create_schema")(CreateSchemaOp)
Operations.register_operation("drop_schema")(DropSchemaOp)


@comparators.dispatch_for("schema")
def compare_schemas(autogen_context, upgrade_ops, schemas: Schemas):
    schemas = autogen_context.metadata.info.get("schemas")
    if not schemas:
        return

    result = schema.compare.compare_schemas(autogen_context.connection, schemas)
    upgrade_ops.ops.extend(result)


@renderers.dispatch_for(CreateSchemaOp)
def render_create_schema(_, op: CreateSchemaOp):
    return "op.create_schema('{}')".format(op.schema.name)


@renderers.dispatch_for(DropSchemaOp)
def render_drop_schema(_, op: DropSchemaOp):
    return "op.drop_schema('{}')".format(op.schema.name)


@Operations.implementation_for(CreateSchemaOp)
def create_schema(operations, operation: CreateSchemaOp):
    operations.execute(CreateSchema(operation.schema.name))


@Operations.implementation_for(DropSchemaOp)
def drop_schema(operations, operation: DropSchemaOp):
    operations.execute(DropSchema(operation.schema.name))



from typing import List, TypeVar

from sqlalchemy_declarative_extensions.dialects.from_string import (
    FromStrings,
    FromStringsSelf,
)


class GrantOptions(FromStrings):
    def default(self: FromStringsSelf) -> List[FromStringsSelf]:
        """Return the default grants given by postgres for the current grant kind.

        PostgreSQL grants default privileges on some types of objects to PUBLIC.
        No privileges are granted to PUBLIC by default on tables, columns,
        schemas or tablespaces.
        """

        return []


class DatabaseGrants(GrantOptions):
    create = "CREATE"
    connect = "CONNECT"
    temporary = "TEMPORARY"

    def default(self):
        return [self.connect, self.temporary]

    @classmethod
    def acl_symbols(cls):
        return {
            cls.create: "C",
            cls.connect: "c",
            cls.temporary: "T",
        }


class ForeignDataWrapperGrants(GrantOptions):
    usage = "USAGE"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.usage: "U",
        }


class ForeignServerGrants(GrantOptions):
    usage = "USAGE"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.usage: "U",
        }


class ForeignTableGrants(GrantOptions):
    select = "SELECT"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.select: "s",
        }


class FunctionGrants(GrantOptions):
    execute = "EXECUTE"

    def default(self):
        return [self.execute]

    @classmethod
    def acl_symbols(cls):
        return {
            cls.execute: "X",
        }


class LanguageGrants(GrantOptions):
    usage = "USAGE"

    def default(self):
        return [self.usage]

    @classmethod
    def acl_symbols(cls):
        return {
            cls.usage: "U",
        }


class LargeObjectGrants(GrantOptions):
    select = "SELECT"
    update = "UPDATE"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.select: "r",
            cls.update: "w",
        }


class TableGrants(GrantOptions):
    select = "SELECT"
    insert = "INSERT"
    update = "UPDATE"
    delete = "DELETE"
    truncate = "TRUNCATE"
    references = "REFERENCES"
    trigger = "TRIGGER"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.select: "r",
            cls.update: "w",
            cls.insert: "a",
            cls.references: "x",
            cls.delete: "d",
            cls.trigger: "t",
            cls.truncate: "D",
        }


class TablespaceGrants(GrantOptions):
    create = "CREATE"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.create: "C",
        }


class TypeGrants(GrantOptions):
    usage = "USAGE"

    def default(self):
        return [self.usage]

    @classmethod
    def acl_symbols(cls):
        return {
            cls.usage: "U",
        }


class SchemaGrants(GrantOptions):
    create = "CREATE"
    usage = "USAGE"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.create: "C",
            cls.usage: "U",
        }


G = TypeVar(
    "G",
    DatabaseGrants,
    ForeignDataWrapperGrants,
    ForeignServerGrants,
    ForeignTableGrants,
    FunctionGrants,
    LanguageGrants,
    LargeObjectGrants,
    SchemaGrants,
    TableGrants,
    TablespaceGrants,
    TypeGrants,
)


class SequenceGrants(GrantOptions):
    usage = "USAGE"
    select = "SELECT"
    update = "UPDATE"

    @classmethod
    def acl_symbols(cls):
        return {
            cls.select: "r",
            cls.update: "w",
            cls.usage: "U",
        }


class GrantTypes(FromStrings):
    database = "DATABASE"
    foreign_data_wrapper = "FOREIGN DATA WRAPPER"
    foreign_server = "FOREIGN SERVER"
    foreign_table = "FOREIGN TABLE"
    function = "FUNCTION"
    language = "LANGUAGE"
    large_object = "LARGE OBJECT"
    schema = "SCHEMA"
    sequence = "SEQUENCE"
    table = "TABLE"
    tablespace = "TABLESPACE"
    type = "TYPE"

    @classmethod
    def _str_to_kind(cls):
        return {
            "f": cls.function,
            "n": cls.schema,
            "S": cls.sequence,
            "r": cls.table,
            "T": cls.type,
        }

    @classmethod
    def from_relkind(cls, relkind: str):
        return cls._str_to_kind()[relkind]

    def to_variants(self):
        return {
            self.database: DatabaseGrants,
            self.function: FunctionGrants,
            self.language: LanguageGrants,
            self.schema: SchemaGrants,
            self.sequence: SequenceGrants,
            self.table: TableGrants,
            self.type: TypeGrants,
        }[self]


class DefaultGrantTypes(FromStrings):
    function = "FUNCTION"
    table = "TABLE"
    type = "TYPE"
    sequence = "SEQUENCE"

    @classmethod
    def _str_to_kind(cls):
        return {
            "f": cls.function,
            "r": cls.table,
            "T": cls.type,
            "S": cls.sequence,
        }

    @classmethod
    def from_relkind(cls, relkind: str):
        return cls._str_to_kind()[relkind]

    def to_variants(self):
        return {
            self.table: TableGrants,
            self.sequence: SequenceGrants,
            self.function: FunctionGrants,
            self.type: TypeGrants,
        }[self]

    def to_grant_type(self):
        return {
            self.table: GrantTypes.table,
            self.sequence: GrantTypes.sequence,
            self.function: GrantTypes.function,
            self.type: GrantTypes.type,
        }[self]



from typing import Optional

from alembic.autogenerate.api import AutogenContext
from alembic.autogenerate.compare import comparators
from alembic.autogenerate.render import renderers
from alembic.operations import Operations
from alembic.operations.ops import UpgradeOps

from sqlalchemy_declarative_extensions import row
from sqlalchemy_declarative_extensions.row.base import Rows
from sqlalchemy_declarative_extensions.row.compare import (
    DeleteRowOp,
    InsertRowOp,
    UpdateRowOp,
)

Operations.register_operation("insert_table_row")(InsertRowOp)
Operations.register_operation("update_table_row")(UpdateRowOp)
Operations.register_operation("delete_table_row")(DeleteRowOp)


@comparators.dispatch_for("schema")
def compare_rows(autogen_context: AutogenContext, upgrade_ops: UpgradeOps, _):
    if (
        autogen_context.metadata is None or autogen_context.connection is None
    ):  # pragma: no cover
        return

    rows: Optional[Rows] = autogen_context.metadata.info.get("rows")
    if not rows:
        return

    result = row.compare.compare_rows(
        autogen_context.connection, autogen_context.metadata, rows
    )
    upgrade_ops.ops.extend(result)  # type: ignore


@renderers.dispatch_for(InsertRowOp)
def render_insert_table_row(_, op: InsertRowOp):
    return "op.insert_table_row('{table}', {values})".format(
        table=op.table, values=op.values
    )


@renderers.dispatch_for(UpdateRowOp)
def render_update_table_row(_, op: UpdateRowOp):
    return "op.update_table_row('{}', from_values={}, to_values={})".format(
        op.table, op.from_values, op.to_values
    )


@renderers.dispatch_for(DeleteRowOp)
def render_delete_table_row(_, op: DeleteRowOp):
    return "op.delete_table_row('{table}', {values})".format(
        table=op.table, values=op.values
    )


@Operations.implementation_for(InsertRowOp)
def insert_row(operations: Operations, operation: InsertRowOp):
    metadata = operations.migration_context.opts["target_metadata"]
    query = operation.render(metadata)
    conn = operations.get_bind()
    conn.execute(query)


@Operations.implementation_for(UpdateRowOp)
def update_row(operations: Operations, operation: UpdateRowOp):
    metadata = operations.migration_context.opts["target_metadata"]
    query = operation.render(metadata)
    conn = operations.get_bind()
    conn.execute(query)


@Operations.implementation_for(DeleteRowOp)
def delete_row(operations: Operations, operation: DeleteRowOp):
    metadata = operations.migration_context.opts["target_metadata"]
    query = operation.render(metadata)
    conn = operations.get_bind()
    conn.execute(query)



from sqlalchemy_declarative_extensions.dialects.postgresql.grant import (
    DefaultGrant,
    DefaultGrantStatement,
    Grant,
    GrantStatement,
)
from sqlalchemy_declarative_extensions.dialects.postgresql.grant_type import (
    DefaultGrantTypes,
    FunctionGrants,
    GrantTypes,
    SchemaGrants,
    SequenceGrants,
    TableGrants,
    TypeGrants,
)
from sqlalchemy_declarative_extensions.dialects.postgresql.role import Role

__all__ = [
    "TableGrants",
    "SequenceGrants",
    "FunctionGrants",
    "SchemaGrants",
    "TypeGrants",
    "GrantTypes",
    "DefaultGrantTypes",
    "Grant",
    "DefaultGrant",
    "DefaultGrantStatement",
    "GrantStatement",
    "DefaultGrant",
    "DefaultGrantStatement",
    "Grant",
    "GrantStatement",
    "Role",
]



from sqlalchemy import MetaData
from sqlalchemy.engine import Connection

from sqlalchemy_declarative_extensions.row import Rows
from sqlalchemy_declarative_extensions.row.compare import compare_rows


def rows_query(rows: Rows):
    def receive_after_create(metadata: MetaData, connection: Connection, **_):
        results = compare_rows(connection, metadata, rows)
        for result in results:
            query = result.render(metadata)
            connection.execute(query)

    return receive_after_create



from alembic.autogenerate.api import AutogenContext
from alembic.autogenerate.compare import comparators
from alembic.autogenerate.render import renderers
from alembic.operations import Operations

from sqlalchemy_declarative_extensions.role.compare import (
    compare_roles,
    CreateRoleOp,
    DropRoleOp,
    UpdateRoleOp,
)

Operations.register_operation("create_role")(CreateRoleOp)
Operations.register_operation("update_role")(UpdateRoleOp)
Operations.register_operation("drop_role")(DropRoleOp)


@comparators.dispatch_for("schema")
def _compare_roles(autogen_context, upgrade_ops, _):
    roles = autogen_context.metadata.info.get("roles")
    if not roles:
        return

    result = compare_roles(autogen_context.connection, roles)
    upgrade_ops.ops[0:0] = result


@renderers.dispatch_for(CreateRoleOp)
def render_create_role(autogen_context: AutogenContext, op: CreateRoleOp):
    command = op.role.to_sql_create()
    return f'op.execute("""{command}""")'


@renderers.dispatch_for(UpdateRoleOp)
def render_update_role(autogen_context: AutogenContext, op: UpdateRoleOp):
    commands = op.from_role.to_sql_update(op.to_role)
    return [f'op.execute("""{command}""")' for command in commands]


@renderers.dispatch_for(DropRoleOp)
def render_drop_role(autogen_context: AutogenContext, op: DropRoleOp):
    command = op.role.to_sql_drop()
    return f'op.execute("""{command}""")'


@Operations.implementation_for(CreateRoleOp)
def create_role(operations, op):
    command = op.role.to_sql_create()
    operations.execute(command)


@Operations.implementation_for(UpdateRoleOp)
def update_role(operations, op):
    commands = op.from_role.to_sql_update(op.to_role)
    for command in commands:
        operations.execute(command)


@Operations.implementation_for(DropRoleOp)
def drop_role(operations, op):
    command = op.role.to_sql_drop()
    operations.execute(command)



import enum
import functools
from typing import Iterable, List, Type, TypeVar, Union

FromStringsSelf = TypeVar("FromStringsSelf", bound="FromStrings")


@functools.total_ordering
class FromStrings(enum.Enum):
    @classmethod
    def from_strings(
        cls: Type[FromStringsSelf], strings: Iterable[Union[str, FromStringsSelf]]
    ) -> List[FromStringsSelf]:
        return [cls.from_string(string) for string in strings]

    @classmethod
    def from_string(
        cls: Type[FromStringsSelf], string: Union[str, FromStringsSelf]
    ) -> FromStringsSelf:
        if isinstance(string, str):
            normalized_str = string.upper()
            return cls(normalized_str)
        else:
            return string

    def __lt__(self, other):
        return self.value < other.value



"""Abstract a postgresql GRANT statement.

See https://www.postgresql.org/docs/latest/sql-grant.html.

ALTER DEFAULT PRIVILEGES
    [ FOR { ROLE | USER } target_role [, ...] ]
    [ IN SCHEMA schema_name [, ...] ]
    abbreviated_grant_or_revoke

where abbreviated_grant_or_revoke is one of:

GRANT { { SELECT | INSERT | UPDATE | DELETE | TRUNCATE | REFERENCES | TRIGGER }
    [, ...] | ALL [ PRIVILEGES ] }
    ON TABLES
    TO { [ GROUP ] role_name | PUBLIC } [, ...] [ WITH GRANT OPTION ]

GRANT { { USAGE | SELECT | UPDATE }
    [, ...] | ALL [ PRIVILEGES ] }
    ON SEQUENCES
    TO { [ GROUP ] role_name | PUBLIC } [, ...] [ WITH GRANT OPTION ]

GRANT { EXECUTE | ALL [ PRIVILEGES ] }
    ON FUNCTIONS
    TO { [ GROUP ] role_name | PUBLIC } [, ...] [ WITH GRANT OPTION ]

GRANT { USAGE | ALL [ PRIVILEGES ] }
    ON TYPES
    TO { [ GROUP ] role_name | PUBLIC } [, ...] [ WITH GRANT OPTION ]
"""
from __future__ import annotations

import itertools
from dataclasses import dataclass, replace
from typing import Generic, List, Optional, Tuple, Union

from sqlalchemy.sql.elements import TextClause
from sqlalchemy.sql.expression import text
from typing_extensions import Protocol, runtime_checkable

from sqlalchemy_declarative_extensions.dialects.postgresql.grant_type import (
    DefaultGrantTypes,
    G,
    GrantOptions,
    GrantTypes,
)


@runtime_checkable
class HasName(Protocol):
    name: str


@dataclass(frozen=True)
class Grant(Generic[G]):
    grants: Tuple[Union[str, G], ...]
    target_role: str
    grant_option: bool = False
    revoke_: bool = False

    @classmethod
    def new(
        cls,
        grant: Union[str, G],
        *grants: Union[str, G],
        to: Union[str, HasName],
        grant_option=False,
    ) -> Grant:
        return cls(
            grants=tuple(sorted([grant, *grants])),
            target_role=_coerce_name(to),
            grant_option=grant_option,
        )

    def revoke(self) -> Grant:
        return replace(self, revoke_=True)

    def with_grant_option(self):
        return replace(self, grant_option=True)

    def on_objects(self, *objects: Union[str, HasName], object_type: GrantTypes):
        variants = object_type.to_variants()
        grant = replace(self, grants=tuple(_map_grant_names(variants, *self.grants)))

        names = [_coerce_name(obj) for obj in objects]
        return GrantStatement(grant, grant_type=object_type, targets=tuple(names))

    def on_tables(self, *tables: Union[str, HasName]):
        return self.on_objects(*tables, object_type=GrantTypes.table)

    def on_sequences(
        self,
        *sequences: Union[str, HasName],
    ):
        return self.on_objects(*sequences, object_type=GrantTypes.sequence)

    def on_schemas(self, *schemas: Union[str, HasName]):
        return self.on_objects(*schemas, object_type=GrantTypes.schema)


@dataclass(frozen=True)
class DefaultGrant:
    grant_type: DefaultGrantTypes
    in_schemas: Tuple[str, ...]
    target_role: Optional[str] = None

    @classmethod
    def on_tables_in_schema(
        cls, *in_schemas: Union[str, HasName], for_role: Optional[HasName] = None
    ) -> DefaultGrant:
        schemas = _map_schema_names(*in_schemas)
        return cls(
            grant_type=DefaultGrantTypes.table,
            in_schemas=tuple(schemas),
            target_role=_coerce_name(for_role) if for_role is not None else None,
        )

    @classmethod
    def on_sequences_in_schema(
        cls, *in_schemas: Union[str, HasName], for_role: Optional[HasName] = None
    ) -> DefaultGrant:
        schemas = _map_schema_names(*in_schemas)
        return cls(
            grant_type=DefaultGrantTypes.sequence,
            in_schemas=tuple(schemas),
            target_role=_coerce_name(for_role) if for_role is not None else None,
        )

    @classmethod
    def on_types_in_schema(
        cls, *in_schemas: Union[str, HasName], for_role: Optional[HasName] = None
    ) -> DefaultGrant:
        schemas = _map_schema_names(*in_schemas)
        return cls(
            grant_type=DefaultGrantTypes.type,
            in_schemas=tuple(schemas),
            target_role=_coerce_name(for_role) if for_role is not None else None,
        )

    @classmethod
    def on_functions_in_schema(
        cls, *in_schemas: Union[str, HasName], for_role: Optional[HasName] = None
    ) -> DefaultGrant:
        schemas = _map_schema_names(*in_schemas)
        return cls(
            grant_type=DefaultGrantTypes.function,
            in_schemas=tuple(schemas),
            target_role=_coerce_name(for_role) if for_role is not None else None,
        )

    def for_role(self, role: str):
        return replace(self, target_role=role)

    def grant(
        self,
        grant: Union[str, G, Grant],
        *grants: Union[str, G],
        to,
        grant_option=False,
    ):
        if not isinstance(grant, Grant):
            grant = Grant(
                grants=tuple(
                    _map_grant_names(self.grant_type.to_variants(), grant, *grants)
                ),
                target_role=to,
                grant_option=grant_option,
            )
        return DefaultGrantStatement(self, grant)


@dataclass(frozen=True)
class DefaultGrantStatement(Generic[G]):
    default_grant: DefaultGrant
    grant: Grant[G]

    def for_role(self, role: Union[str, HasName]) -> DefaultGrantStatement:
        return replace(
            self, default_grant=replace(self.default_grant, target_role=role)
        )

    def invert(self) -> DefaultGrantStatement:
        return replace(self, grant=replace(self.grant, revoke_=not self.grant.revoke_))

    def to_sql(self) -> TextClause:
        result = []

        result.append("ALTER DEFAULT PRIVILEGES")

        if self.default_grant.target_role:
            result.append(f'FOR ROLE "{self.default_grant.target_role}"')

        schemas_str = ", ".join([f'"{t}"' for t in self.default_grant.in_schemas])
        result.append(f"IN SCHEMA {schemas_str}")

        result.append(_render_grant_or_revoke(self.grant))
        result.append(_render_privilege(self.grant, self.default_grant.grant_type))

        result.append(f"ON {self.default_grant.grant_type.value}S")
        result.append(_render_to_or_from(self.grant))

        text_result = " ".join(result)
        return text(text_result + ";")

    def explode(self):
        return [
            DefaultGrantStatement(
                default_grant=DefaultGrant(
                    grant_type=self.default_grant.grant_type,
                    in_schemas=(schema,),
                    target_role=self.default_grant.target_role,
                ),
                grant=Grant(
                    grants=(grant,),
                    target_role=self.grant.target_role,
                    grant_option=self.grant.grant_option,
                    revoke_=self.grant.revoke_,
                ),
            )
            for schema in self.default_grant.in_schemas
            for grant in self.grant.grants
        ]

    @classmethod
    def combine(cls, grants: List[DefaultGrantStatement]):
        def by_statement(g: DefaultGrantStatement):
            return (
                g.default_grant.grant_type,
                g.default_grant.in_schemas,
                g.default_grant.target_role or "",
                g.grant.target_role,
                g.grant.grant_option,
                g.grant.revoke_,
            )

        result = []
        groups = itertools.groupby(sorted(grants, key=by_statement), key=by_statement)
        for (
            grant_type,
            in_schemas,
            default_target_role,
            target_role,
            grant_option,
            revoke,
        ), group in groups:
            item = cls(
                default_grant=DefaultGrant(
                    grant_type=grant_type,
                    in_schemas=in_schemas,
                    target_role=default_target_role or None,
                ),
                grant=Grant(
                    target_role=target_role,
                    grant_option=grant_option,
                    revoke_=revoke,
                    grants=tuple([g for i in group for g in i.grant.grants]),
                ),
            )
            result.append(item)
        return result


@dataclass(frozen=True)
class GrantStatement(Generic[G]):
    grant: Grant[G]
    grant_type: GrantTypes
    targets: Tuple[str, ...]

    def invert(self) -> GrantStatement:
        return replace(self, grant=replace(self.grant, revoke_=not self.grant.revoke_))

    def for_role(self, role: Union[str, HasName]) -> GrantStatement:
        return replace(self, grant=replace(self.grant, target_role=_coerce_name(role)))

    def to_sql(self) -> TextClause:
        result = []

        result.append(_render_grant_or_revoke(self.grant))
        result.append(_render_privilege(self.grant, self.grant_type))
        result.append(f"ON {self.grant_type.value}")

        result.append(", ".join([_quote_table_name(t) for t in self.targets]))
        result.append(_render_to_or_from(self.grant))

        grant_option = _render_grant_option(self.grant)
        if grant_option:
            result.append(grant_option)

        text_result = " ".join(result)
        return text(text_result + ";")

    def explode(self):
        return [
            GrantStatement(
                grant=Grant(
                    grants=(grant,),
                    target_role=self.grant.target_role,
                    grant_option=self.grant.grant_option,
                    revoke_=self.grant.revoke_,
                ),
                grant_type=self.grant_type,
                targets=(target,),
            )
            for target in self.targets
            for grant in self.grant.grants
        ]

    @classmethod
    def combine(cls, grants: List[GrantStatement]):
        def by_statement(g: GrantStatement):
            return (
                g.grant_type,
                g.targets,
                g.grant.target_role,
                g.grant.grant_option,
                g.grant.revoke_,
            )

        result = []
        groups = itertools.groupby(sorted(grants, key=by_statement), key=by_statement)
        for (grant_type, targets, target_role, grant_option, revoke), group in groups:
            item = cls(
                grant_type=grant_type,
                targets=targets,
                grant=Grant(
                    target_role=target_role,
                    grant_option=grant_option,
                    revoke_=revoke,
                    grants=tuple([g for i in group for g in i.grant.grants]),
                ),
            )
            result.append(item)
        return result


def _render_grant_or_revoke(grant: Grant) -> str:
    if grant.revoke_:
        return "REVOKE"
    return "GRANT"


def _render_to_or_from(grant: Grant) -> str:
    if grant.revoke_:
        return f'FROM "{grant.target_role}"'
    return f'TO "{grant.target_role}"'


def _quote_table_name(name: str):
    if "." in name:
        schema, name = name.split(".")
    else:
        schema = None

    if schema:
        return f'"{schema}"."{name}"'
    return f'"{name}"'


def _render_privilege(
    grant: Grant, grant_type: Union[DefaultGrantTypes, GrantTypes]
) -> str:
    grant_variant_cls = grant_type.to_variants()
    return ", ".join(v.value for v in grant_variant_cls.from_strings(grant.grants))


def _render_grant_option(grant: Grant) -> Optional[str]:
    if grant.grant_option:
        return "WITH GRANT OPTION"
    return None


def _map_schema_names(*schemas: Union[str, HasName]):
    return sorted([_coerce_name(s) for s in schemas])


def _map_grant_names(variant: G, *grants: Union[str, G]):
    return sorted(
        [g if isinstance(g, GrantOptions) else variant.from_string(g) for g in grants]
    )


def _coerce_name(name: Union[str, HasName]):
    if isinstance(name, HasName):
        return name.name
    return name



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Iterable, Optional, Sequence, Union


@dataclass(frozen=True)
class Schemas:
    """A collection of schemas and the settings for diff/collection.

    Arguments:
        schemas: The list of grants
        ignore_unspecified: Optionally ignore detected grants which do not match
            the set of defined grants.

    Examples:
        - No schemas

        >>> schemas = Schemas()

        - Some options set

        >>> schemas = Schemas(ignore_unspecified=True)

        - With some actual schemas

        >>> from sqlalchemy_declarative_extensions import Schema, Schemas
        >>> schema = Schemas().are("foo", Schema("bar"), ...)
    """

    schemas: Sequence[Schema] = ()
    ignore_unspecified: bool = False

    @classmethod
    def coerce_from_unknown(
        cls, unknown: Union[None, Iterable[Union[Schema, str]], Schemas]
    ) -> Optional[Schemas]:
        if isinstance(unknown, Schemas):
            return unknown

        if isinstance(unknown, Iterable):
            return cls().are(*unknown)

        return None

    def __iter__(self):
        for schema in self.schemas:
            yield schema

    def are(self, *schemas: Union[Schema, str]):
        """Declare the set of schemas which should exist."""
        result = replace(
            self,
            schemas=tuple([Schema.coerce_from_unknown(schema) for schema in schemas]),
        )
        return result


@dataclass(frozen=True, order=True)
class Schema:
    """Represents a schema."""

    name: str

    @classmethod
    def coerce_from_unknown(cls, unknown: Union[Schema, str]) -> Schema:
        if isinstance(unknown, Schema):
            return unknown

        return cls(unknown)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Union

from sqlalchemy import MetaData, Table, tuple_
from sqlalchemy.engine.base import Connection

from sqlalchemy_declarative_extensions.row.base import Row, Rows


@dataclass
class InsertRowOp:
    table: str
    values: dict[str, Any]

    @classmethod
    def insert_table_row(cls, operations, table, values):
        op = cls(table, values)
        return operations.invoke(op)

    def render(self, metadata: MetaData):
        table = metadata.tables[self.table]
        return table.insert().values(self.values)

    def reverse(self):
        return DeleteRowOp(self.table, self.values)


@dataclass
class UpdateRowOp:
    table: str
    from_values: dict[str, Any]
    to_values: dict[str, Any]

    @classmethod
    def update_table_row(cls, operations, table, from_values, to_values):
        op = cls(table, from_values, to_values)
        return operations.invoke(op)

    def render(self, metadata: MetaData):
        table = metadata.tables[self.table]
        primary_key_columns = [c.name for c in table.primary_key.columns]
        where = [
            table.c[c] == v
            for c, v in self.to_values.items()
            if c in primary_key_columns
        ]
        values = {
            c: v for c, v in self.to_values.items() if c not in primary_key_columns
        }
        return table.update().where(*where).values(**values)

    def reverse(self):
        return UpdateRowOp(self.table, self.to_values, self.from_values)


@dataclass
class DeleteRowOp:
    table: str
    values: dict[str, Any]

    @classmethod
    def delete_table_row(cls, operations, table, values):
        op = cls(table, values)
        return operations.invoke(op)

    def render(self, metadata: MetaData):
        table = metadata.tables[self.table]
        primary_key_columns = [c.name for c in table.primary_key.columns]
        where = [
            table.c[c] == v for c, v in self.values.items() if c in primary_key_columns
        ]
        return table.delete().where(*where)

    def reverse(self):
        return InsertRowOp(self.table, self.values)


RowOp = Union[InsertRowOp, UpdateRowOp, DeleteRowOp]


def compare_rows(connection: Connection, metadata: MetaData, rows: Rows) -> List[RowOp]:
    result: List[RowOp] = []

    rows_by_table: Dict[Table, List[Row]] = {}
    for row in rows:
        table = metadata.tables.get(row.qualified_name)
        if table is None:
            raise ValueError(f"Unknown table: {row.qualified_name}")

        rows_by_table.setdefault(table, []).append(row)

        primary_key_columns = [c.name for c in table.primary_key.columns]

        if set(primary_key_columns) - row.column_values.keys():
            raise ValueError(
                f"Row is missing primary key values required to declaratively specify: {row}"
            )

        column_filters = [
            c == row.column_values[c.name] for c in table.primary_key.columns
        ]
        record = connection.execute(
            table.select().where(*column_filters).limit(1)
        ).first()
        if record:
            row_keys = row.column_values.keys()
            record_dict = {k: v for k, v in dict(record).items() if k in row_keys}
            if row.column_values == record_dict:
                continue
            else:
                result.append(
                    UpdateRowOp(
                        row.qualified_name,
                        from_values=record_dict,
                        to_values=row.column_values,
                    )
                )
        else:
            result.append(InsertRowOp(row.qualified_name, values=row.column_values))

    if not rows.ignore_unspecified:
        for table_name in rows.included_tables:
            table = metadata.tables[table_name]
            rows_by_table.setdefault(table, [])

        for table, row_list in rows_by_table.items():
            primary_key_columns = [c.name for c in table.primary_key.columns]
            primary_key_values = [
                tuple(row.column_values[c] for c in primary_key_columns)
                for row in row_list
            ]
            to_delete = connection.execute(
                table.select().where(
                    tuple_(*table.primary_key.columns).notin_(primary_key_values)
                )
            ).all()

            for record in to_delete:
                op = DeleteRowOp(table.fullname, dict(record))
                result.append(op)

    return result



from typing import Optional

from sqlalchemy import MetaData
from sqlalchemy.engine import Connection

from sqlalchemy_declarative_extensions import Roles
from sqlalchemy_declarative_extensions.role.compare import compare_roles


def role_ddl(metadata: MetaData, connection: Connection, **_):
    roles: Optional[Roles] = metadata.info.get("roles")
    if not roles:  # pragma: no cover
        return

    result = compare_roles(connection, roles)
    for op in result:
        connection.execute(op.to_sql())



from sqlalchemy.schema import CreateSchema

from sqlalchemy_declarative_extensions.dialects import check_schema_exists
from sqlalchemy_declarative_extensions.schema import Schema


def schema_ddl(schema: Schema):
    ddl = CreateSchema(schema.name)
    return ddl.execute_if(callable_=check_schema)  # type: ignore


def check_schema(ddl, target, connection, **_):
    schema = ddl.element
    return check_schema_exists(connection, name=schema)



from sqlalchemy import and_, column, literal, select, String, table, text, union
from sqlalchemy.dialects.postgresql import ARRAY

pg_class = table(
    "pg_class",
    column("oid"),
    column("relname"),
    column("relnamespace"),
    column("relacl"),
    column("relkind"),
    column("relowner"),
)

pg_namespace = table(
    "pg_namespace",
    column("oid"),
    column("nspname"),
    column("nspowner"),
    column("nspacl"),
)

pg_roles = table(
    "pg_roles",
    column("oid"),
    column("rolname"),
)

pg_default_acl = table(
    "pg_default_acl",
    column("defaclrole"),
    column("defaclnamespace"),
    column("defaclobjtype"),
    column("defaclacl"),
)

pg_authid = table(
    "pg_authid",
    column("oid"),
    column("rolname"),
)


roles_query = text(
    """
        SELECT r.rolname, r.rolsuper, r.rolinherit,
          r.rolcreaterole, r.rolcreatedb, r.rolcanlogin,
          r.rolconnlimit, r.rolvaliduntil,
          ARRAY(SELECT b.rolname
                FROM pg_catalog.pg_auth_members m
                JOIN pg_catalog.pg_roles b ON (m.roleid = b.oid)
                WHERE m.member = r.oid) as memberof
        , r.rolreplication
        , r.rolbypassrls
        FROM pg_catalog.pg_roles r
        WHERE r.rolname !~ '^pg_'
        ORDER BY 1;
        """
)

_schema_not_pg = and_(
    pg_namespace.c.nspname != "information_schema",
    pg_namespace.c.nspname.not_like("pg_%"),
)
_schema_not_public = pg_namespace.c.nspname != "public"
_table_not_pg = pg_class.c.relname.not_like("pg_%")

schemas_query = (
    select(pg_namespace.c.nspname).where(_schema_not_pg).where(_schema_not_public)
)


schema_exists_query = text(
    "SELECT schema_name FROM information_schema.schemata WHERE schema_name = :schema"
)


default_acl_query = select(
    pg_authid.c.rolname.label("role_name"),
    pg_namespace.c.nspname.label("schema_name"),
    pg_default_acl.c.defaclobjtype.label("object_type"),
    pg_default_acl.c.defaclacl.cast(ARRAY(String)).label("acl"),
).select_from(
    pg_default_acl.join(pg_authid, pg_default_acl.c.defaclrole == pg_authid.c.oid).join(
        pg_namespace, pg_default_acl.c.defaclnamespace == pg_namespace.c.oid
    )
)

object_acl_query = union(
    select(
        pg_namespace.c.nspname.label("schema"),
        pg_class.c.relname.label("name"),
        pg_class.c.relkind.label("relkind"),
        pg_authid.c.rolname.label("owner"),
        pg_class.c.relacl.cast(ARRAY(String)).label("acl"),
    )
    .select_from(
        pg_class.join(pg_namespace, pg_class.c.relnamespace == pg_namespace.c.oid).join(
            pg_authid, pg_class.c.relowner == pg_authid.c.oid
        )
    )
    .where(pg_class.c.relkind.in_(["r", "S", "f", "n", "T"]))
    .where(_table_not_pg)
    .where(_schema_not_pg),
    select(
        literal(None).label("schema"),
        pg_namespace.c.nspname.label("name"),
        literal("n").label("relkind"),
        pg_authid.c.rolname.label("owner"),
        pg_namespace.c.nspacl.cast(ARRAY(String)),
    )
    .select_from(
        pg_namespace.join(pg_authid, pg_namespace.c.nspowner == pg_authid.c.oid)
    )
    .where(_schema_not_pg)
    .where(_schema_not_public),
)

objects_query = (
    select(
        pg_namespace.c.nspname.label("schema"),
        pg_class.c.relname.label("object_name"),
        pg_class.c.relkind.label("relkind"),
    )
    .select_from(
        pg_class.join(pg_namespace, pg_class.c.relnamespace == pg_namespace.c.oid)
    )
    .where(pg_class.c.relkind.in_(["r", "S", "f", "n", "T"]))
    .where(_table_not_pg)
    .where(_schema_not_pg)
)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

from sqlalchemy.engine.base import Connection

from sqlalchemy_declarative_extensions.dialects import get_schemas
from sqlalchemy_declarative_extensions.schema.base import Schema, Schemas


@dataclass
class CreateSchemaOp:
    schema: Schema

    @classmethod
    def create_schema(cls, operations, schema, **kwargs):
        op = cls(Schema(schema, **kwargs))
        return operations.invoke(op)

    def reverse(self):
        return DropSchemaOp(self.schema)


@dataclass
class DropSchemaOp:
    schema: Schema

    @classmethod
    def drop_schema(cls, operations, schema, **kwargs):
        op = cls(Schema(schema, **kwargs))
        return operations.invoke(op)

    def reverse(self):
        return CreateSchemaOp(self.schema)


SchemaOp = Union[CreateSchemaOp, DropSchemaOp]


def compare_schemas(connection: Connection, schemas: Schemas) -> List[SchemaOp]:
    existing_schemas = get_schemas(connection)

    expected_schemas = set(schemas.schemas)
    new_schemas = expected_schemas - existing_schemas
    removed_schemas = existing_schemas - expected_schemas

    result: List[SchemaOp] = []
    for schema in sorted(new_schemas):
        result.insert(0, CreateSchemaOp(schema))

    if not schemas.ignore_unspecified:
        for schema in reversed(sorted(removed_schemas)):
            result.append(DropSchemaOp(schema))

    return result



from typing import Dict, Iterable, Set

from sqlalchemy_declarative_extensions.role.generic import Role


def topological_sort(roles: Iterable[Role]):
    """Sort roles in an order that guarantees success based on role dependence.

    Uses Kahn's algorithm.
    """
    role_name_map = generate_role_map(roles)
    role_dep_map = generate_role_dependency_map(role_name_map)

    fullfilled_role_names = deduplicate_roles(
        role_dep_map.items(), exclude_no_deps=True
    )

    for role in fullfilled_role_names:
        role_dep_map.pop(role)

    result = []
    while fullfilled_role_names:
        fullfilled_role_name = fullfilled_role_names.pop(0)
        result.append(role_name_map[fullfilled_role_name])

        # Remove fullfilled role from deps of all other roles which might depend on it.
        for deps in role_dep_map.values():
            if fullfilled_role_name in deps:
                deps.remove(fullfilled_role_name)

        newly_fullfilled_roles = deduplicate_roles(
            role_dep_map.items(), exclude_no_deps=True
        )

        fullfilled_role_names = deduplicate_roles(
            fullfilled_role_names + newly_fullfilled_roles
        )
        for role in newly_fullfilled_roles:
            role_dep_map.pop(role)

    if any(role_dep_map):
        cyclical_roles = ", ".join(sorted(role_dep_map.keys()))
        raise ValueError(
            f"The following roles have cyclical dependencies: {cyclical_roles}"
        )

    return result


def deduplicate_roles(deps, exclude_no_deps=False):
    if exclude_no_deps:
        deps = (role for role, deps in deps if not deps)

    return list(dict.fromkeys(deps))


def generate_role_map(roles: Iterable[Role]) -> Dict[str, Role]:
    result: Dict[str, Role] = {}
    for role in roles:
        if role.name in result:
            raise ValueError(f"Duplicate role specified: {role}")
        result[role.name] = role
    return result


def generate_role_dependency_map(role_name_map: Dict[str, Role]) -> Dict[str, Set[str]]:
    result: Dict[str, Set[str]] = {}
    valid_role_names = set(role_name_map)
    for name, role in role_name_map.items():
        in_roles = set(role.in_roles or [])

        if not in_roles.issubset(valid_role_names):
            missing_roles = ", ".join(sorted(in_roles - valid_role_names))
            raise ValueError(
                "The following roles are specified as dependencies of other "
                "top-level roles, but are missing in the list of top-level "
                f"roles: {missing_roles}."
            )
        result[name] = in_roles
    return result



from __future__ import annotations

from typing import Any, Callable, Iterable, Optional, Type, TYPE_CHECKING, Union

from sqlalchemy import event
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql.schema import MetaData
from typing_extensions import Protocol

from sqlalchemy_declarative_extensions.grant.base import Grants
from sqlalchemy_declarative_extensions.role.base import Roles
from sqlalchemy_declarative_extensions.row.base import Row, Rows
from sqlalchemy_declarative_extensions.schema.base import Schemas

if TYPE_CHECKING:
    from sqlalchemy_declarative_extensions.dialects import postgresql
    from sqlalchemy_declarative_extensions.grant.base import G
    from sqlalchemy_declarative_extensions.role import generic
    from sqlalchemy_declarative_extensions.schema.base import Schema


def declarative_database(
    base: Optional[Any] = None,
) -> Union[
    Type[DeclarativeMeta], Callable[[Type[DeclarativeMeta]], Type[DeclarativeMeta]]
]:
    """Decorate a SQLAlchemy declarative base object to register database objects.

    See also :func:`sqlalchemy_declarative_extensions.declare_database`, of which this decorator is syntactic sugar.

    By decorating your declarative base with this decorator, the attributes
    "schemas", "roles", and "grants" are read and will register with SQLAlchemy
    and/or Alembic to ensure they're created during a ``metadata.create_createall``
    call or ``alembic --autogenerate``.

    Examples:
        >>> from sqlalchemy.orm import declarative_base
        >>> Base_ = declarative_base()
        >>>
        >>> from sqlalchemy_declarative_extensions import declarative_database, Roles
        >>> from sqlalchemy_declarative_extensions.dialects.postgresql import Role
        >>> @declarative_database
        ... class Base(Base_):
        ...     __abstract__ = True
        ...
        ...     schemas = Schemas().are('example')
        ...     roles = (
        ...         Roles(ignore_unspecified=True)
        ...         .are(
        ...             'read',
        ...             Role('fancy', createrole=True, in_roles=['read']),
        ...         )
        ...     )

    """

    def _declare_database(cls: Type[DeclarativeMeta]) -> Type[DeclarativeMeta]:
        raw_roles = getattr(cls, "roles", None)
        raw_schemas = getattr(cls, "schemas", None)
        raw_grants = getattr(cls, "grants", None)
        raw_rows = getattr(cls, "rows", None)

        declare_database(
            cls.metadata,
            schemas=raw_schemas,
            roles=raw_roles,
            grants=raw_grants,
            rows=raw_rows,
        )
        return cls

    if base is None:
        return _declare_database
    return _declare_database(base)


def declare_database(
    metadata: MetaData,
    *,
    schemas: Union[None, Iterable[Union[Schema, str]], Schemas] = None,
    roles: Union[
        None, Iterable[Union[generic.Role, postgresql.Role, str]], Roles
    ] = None,
    grants: Union[None, Iterable[G], Grants] = None,
    rows: Union[None, Iterable[Row], Rows] = None,
):
    """Register declaratively specified database extension handlers.

    See also :func:`declarative_database` for alternative API which decorates
    a decalarative base class definition.

    This attaches the given declarative extensions onto the metadata such that
    they'll be recognized by SQLAlchemy's ``metadata.create_all`` call, as well
    as ``alembic --autogenerate`` (if you :func:`register_alembic_events`).

    .. note::

        The argument types are intentionally permissive. For simple schemas and
        roles especially, the most obvious and direct interface might literally
        be a list of strings, and as such that is an accepted input. Most documentation
        examples will reference the typed objects to which raw strings are coerced.

        For more complex examples, and especially if you need meta-options like
        ``ignore_unspecified``, you will need to use the objects directly.

    Arguments:
        metadata: The metadata on which the given extensions are being registered.
        schemas: The set of schemas to ensure exist.
        roles: The set of roles to ensure exist.
        grants: The set of grants to ensure are applied to the roles/schemas/tables.
        rows: The set of rows to ensure are applied to the roles/schemas/tables.
    """
    metadata.info["schemas"] = Schemas.coerce_from_unknown(schemas)
    metadata.info["roles"] = Roles.coerce_from_unknown(roles)
    metadata.info["grants"] = Grants.coerce_from_unknown(grants)
    metadata.info["rows"] = Rows.coerce_from_unknown(rows)


class HasMetaData(Protocol):
    metadata: MetaData


def register_sqlalchemy_events(
    maybe_metadata: Union[MetaData, HasMetaData],
    schemas=False,
    roles=False,
    grants=False,
    rows=False,
):
    """Register handlers for supported object types into SQLAlchemy's event system.

    By default all object types are disabled, but each can be individually enabled.
    We assume most execution environments where one is using `MetaData.create_all`
    will be in tests; where roles and grants, in particular, are database-wide
    objects which can cause issues.

    Note this is the opposite of the defaults when registering against SQLAlchemy's
    event system.
    """
    from sqlalchemy_declarative_extensions.grant.ddl import grant_ddl
    from sqlalchemy_declarative_extensions.role.ddl import role_ddl
    from sqlalchemy_declarative_extensions.row.query import rows_query
    from sqlalchemy_declarative_extensions.schema.ddl import schema_ddl

    if isinstance(maybe_metadata, MetaData):
        metadata = maybe_metadata
    else:
        metadata = maybe_metadata.metadata

    concrete_schemas = metadata.info["schemas"]
    concrete_roles = metadata.info["roles"]
    concrete_grants = metadata.info["grants"]
    concrete_rows = metadata.info["rows"]

    if concrete_schemas and schemas:
        for schema in concrete_schemas:
            event.listen(
                metadata,
                "before_create",
                schema_ddl(schema),
            )

    if concrete_roles and roles:
        event.listen(
            metadata,
            "before_create",
            role_ddl,
        )

    if concrete_grants and grants:
        event.listen(
            metadata,
            "before_create",
            grant_ddl(concrete_grants, after=False),
        )

        event.listen(
            metadata,
            "after_create",
            grant_ddl(concrete_grants, after=True),
        )

    if concrete_rows and rows:
        event.listen(
            metadata,
            "after_create",
            rows_query(concrete_rows),
        )



from __future__ import annotations

from dataclasses import dataclass
from itertools import groupby
from typing import Container, List, Optional, Set, Union

from sqlalchemy.engine import Connection

from sqlalchemy_declarative_extensions.dialects import (
    get_default_grants,
    get_grants,
    get_objects,
)
from sqlalchemy_declarative_extensions.dialects.postgresql import (
    DefaultGrantStatement,
    GrantStatement,
    GrantTypes,
)
from sqlalchemy_declarative_extensions.grant.base import Grants
from sqlalchemy_declarative_extensions.role.base import Roles


@dataclass
class GrantPrivilegesOp:
    grant: Union[DefaultGrantStatement, GrantStatement]

    def reverse(self):
        return RevokePrivilegesOp(self.grant.invert())


@dataclass
class RevokePrivilegesOp:
    grant: Union[DefaultGrantStatement, GrantStatement]

    def reverse(self):
        return GrantPrivilegesOp(self.grant.invert())


Operation = Union[GrantPrivilegesOp, RevokePrivilegesOp]


def compare_grants(
    connection: Connection, grants: Grants, roles: Optional[Roles] = None
) -> List[Operation]:
    result: List[Operation] = []

    current_role: str = connection.engine.url.username  # type: ignore

    filtered_roles: Optional[Set[str]] = None
    if grants.only_defined_roles:
        filtered_roles = {r.name for r in (roles or [])}

    default_grant_ops = compare_default_grants(connection, grants, roles=filtered_roles)
    result.extend(default_grant_ops)

    if grants.default_grants_imply_grants:
        grant_ops = compare_object_grants(
            connection, grants, username=current_role, roles=filtered_roles
        )
        result.extend(grant_ops)

    return result


def compare_default_grants(
    connection: Connection,
    grants: Grants,
    roles: Optional[Container[str]] = None,
):
    result: List[Operation] = []

    existing_default_grants = get_default_grants(connection, roles=roles, expanded=True)

    expected_grants = []
    for grant in grants:
        if not isinstance(grant, DefaultGrantStatement):
            continue

        expected_grants.extend(grant.explode())

    missing_grants = set(expected_grants) - set(existing_default_grants)
    extra_grants = set(existing_default_grants) - set(expected_grants)

    if not grants.ignore_unspecified:
        revoke_statements = [extra_grant.invert() for extra_grant in extra_grants]
        for revoke in DefaultGrantStatement.combine(revoke_statements):
            result.append(RevokePrivilegesOp(revoke))

    for grant in DefaultGrantStatement.combine(list(missing_grants)):
        result.append(RevokePrivilegesOp(grant))

    return result


def compare_object_grants(
    connection: Connection,
    grants: Grants,
    username: str,
    roles: Optional[Container[str]] = None,
):
    result: List[Operation] = []

    expected_grants = [g for g in grants if isinstance(g, GrantStatement)]

    existing_tables = get_objects(connection)
    existing_tables_by_schema = {
        s: list(g) for s, g in groupby(existing_tables, lambda r: r[0])
    }
    for grant in grants:
        if not isinstance(grant, DefaultGrantStatement):
            continue

        grant_type = grant.default_grant.grant_type.to_grant_type()

        for schema in grant.default_grant.in_schemas:
            existing_tables_in_schema = existing_tables_by_schema.get(schema)
            if not existing_tables_in_schema:
                continue

            for _, table, relkind in existing_tables_in_schema:
                object_type = GrantTypes.from_relkind(relkind)

                if object_type == grant_type:
                    expected_grants.extend(
                        grant.grant.on_objects(table, object_type=object_type).explode()
                    )

    existing_grants = get_grants(connection, roles=roles, expanded=True)

    if grants.ignore_self_grants:
        existing_grants = [
            g for g in existing_grants if g.grant.target_role != username
        ]

    missing_grants = set(expected_grants) - set(existing_grants)
    extra_grants = set(existing_grants) - set(expected_grants)

    if not grants.ignore_unspecified:
        revoke_statements = [extra_grant.invert() for extra_grant in extra_grants]
        for revoke in GrantStatement.combine(revoke_statements):
            result.append(RevokePrivilegesOp(revoke))

    for grant in GrantStatement.combine(list(missing_grants)):
        result.append(RevokePrivilegesOp(grant))

    return result



from typing import Optional

from sqlalchemy import MetaData
from sqlalchemy.engine import Connection

from sqlalchemy_declarative_extensions.grant.base import Grants
from sqlalchemy_declarative_extensions.grant.compare import compare_grants
from sqlalchemy_declarative_extensions.role.base import Roles


def grant_ddl(grants: Grants, after: bool):
    def receive_event(metadata: MetaData, connection: Connection, **_):
        roles: Optional[Roles] = metadata.info.get("roles")
        result = compare_grants(connection, grants, roles=roles)
        for op in result:
            connection.execute(op.grant.to_sql())

    return receive_event



from sqlalchemy_declarative_extensions.dialects import postgresql
from sqlalchemy_declarative_extensions.dialects.postgresql.query import (
    check_schema_exists_postgresql,
    get_default_grants_postgresql,
    get_grants_postgresql,
    get_objects_postgresql,
    get_roles_postgresql,
    get_schemas_postgresql,
)
from sqlalchemy_declarative_extensions.dialects.sqlite.query import (
    check_schema_exists_sqlite,
)
from sqlalchemy_declarative_extensions.sqlalchemy import dialect_dispatch

get_schemas = dialect_dispatch(
    postgresql=get_schemas_postgresql,
)

check_schema_exists = dialect_dispatch(
    postgresql=check_schema_exists_postgresql,
    sqlite=check_schema_exists_sqlite,
)

get_objects = dialect_dispatch(
    postgresql=get_objects_postgresql,
)

get_default_grants = dialect_dispatch(
    postgresql=get_default_grants_postgresql,
)

get_grants = dialect_dispatch(
    postgresql=get_grants_postgresql,
)

get_roles = dialect_dispatch(
    postgresql=get_roles_postgresql,
)

get_roles = dialect_dispatch(
    postgresql=get_roles_postgresql,
)

get_role_cls = dialect_dispatch(
    postgresql=lambda _: postgresql.Role,
)



from sqlalchemy import text
from sqlalchemy.engine import Connection


def check_schema_exists_sqlite(connection: Connection, name: str) -> bool:
    """Check whether the given schema exists.

    For `sqlalchemy.schema.CreateSchema` to work, we need to first attach
    a :memory: as the given schema name first. Given that they need to be
    created anew for each new connection, we can (hopefully) safely,
    unconditionally attach it and return `False` always.
    """
    schema_exists = "ATTACH DATABASE ':memory:' AS :schema"
    connection.execute(text(schema_exists), schema=name)
    return False



from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import Iterable, List, Optional, Union


@dataclass
class Rows:
    rows: List[Row] = field(default_factory=list)
    included_tables: List[str] = field(default_factory=list)
    ignore_unspecified: bool = False

    @classmethod
    def coerce_from_unknown(
        cls, unknown: Union[None, Iterable[Row], Rows]
    ) -> Optional[Rows]:
        if isinstance(unknown, Rows):
            return unknown

        if isinstance(unknown, Iterable):
            return Rows().are(*unknown)

        return None

    def __iter__(self):
        for role in self.rows:
            yield role

    def are(self, *rows: Row):
        return replace(self, rows=rows)


class Row:
    def __init__(self, tablename, **column_values):
        try:
            schema, table = tablename.split(".", 1)
        except ValueError:
            self.schema = None
            self.tablename = tablename
        else:
            self.schema = schema
            self.tablename = table

        self.qualified_name = tablename
        self.column_values = column_values



def register_alembic_events(
    schemas: bool = True,
    roles: bool = True,
    grants: bool = True,
    rows: bool = True,
):
    """Register handlers into alembic's event system for the supported object types.

    By default all object types are enabled, but each can be individually disabled.

    Note this is the opposite of the defaults when registering against SQLAlchemy's
    event system.
    """
    if schemas:
        import sqlalchemy_declarative_extensions.alembic.schema  # noqa

    if roles:
        import sqlalchemy_declarative_extensions.alembic.role  # noqa

    if grants:
        import sqlalchemy_declarative_extensions.alembic.grant  # noqa

    if rows:
        import sqlalchemy_declarative_extensions.alembic.row  # noqa



from typing import Optional

from alembic.autogenerate.api import AutogenContext
from alembic.autogenerate.compare import comparators
from alembic.autogenerate.render import renderers
from alembic.operations.ops import UpgradeOps

from sqlalchemy_declarative_extensions.grant import compare
from sqlalchemy_declarative_extensions.grant.base import Grants
from sqlalchemy_declarative_extensions.grant.compare import (
    GrantPrivilegesOp,
    RevokePrivilegesOp,
)
from sqlalchemy_declarative_extensions.role.base import Roles
from sqlalchemy_declarative_extensions.role.compare import RoleOp


@comparators.dispatch_for("schema")
def compare_grants(autogen_context: AutogenContext, upgrade_ops: UpgradeOps, _):
    if autogen_context.metadata is None or autogen_context.connection is None:
        return  # pragma: no cover

    grants: Optional[Grants] = autogen_context.metadata.info.get("grants")
    if not grants:
        return

    roles: Optional[Roles] = autogen_context.metadata.info.get("roles")

    result = compare.compare_grants(autogen_context.connection, grants, roles=roles)
    if not result:
        return

    # Find the index of the last element of a role ops, which this needs to be after.
    last_role_index = -1
    for index, op in enumerate(upgrade_ops.ops):
        if isinstance(op, RoleOp):
            last_role_index = index

    # Insert after that point
    after_last_role_index = last_role_index + 1
    upgrade_ops.ops[after_last_role_index:after_last_role_index] = result  # type: ignore


@renderers.dispatch_for(GrantPrivilegesOp)
def render_grant(_, op: RevokePrivilegesOp):
    return f'op.execute(sa.text("""{op.grant.to_sql()}"""))'


@renderers.dispatch_for(RevokePrivilegesOp)
def render_revoke(_, op: RevokePrivilegesOp):
    return f'op.execute(sa.text("""{op.grant.to_sql()}"""))'



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sqlalchemy_declarative_extensions',
 'sqlalchemy_declarative_extensions.alembic',
 'sqlalchemy_declarative_extensions.dialects',
 'sqlalchemy_declarative_extensions.dialects.postgresql',
 'sqlalchemy_declarative_extensions.dialects.sqlite',
 'sqlalchemy_declarative_extensions.grant',
 'sqlalchemy_declarative_extensions.role',
 'sqlalchemy_declarative_extensions.row',
 'sqlalchemy_declarative_extensions.schema']

package_data = \
{'': ['*']}

install_requires = \
['sqlalchemy>=1.3']

extras_require = \
{':python_version < "3.11"': ['typing_extensions>=4.0']}

setup_kwargs = {
    'name': 'sqlalchemy-declarative-extensions',
    'version': '0.2.2',
    'description': 'Library to declare additional kinds of objects not natively supported by SqlAlchemy/Alembic.',
    'long_description': '# SqlAlchemy Declarative Extensions\n\n[![Actions Status](https://github.com/dancardin/sqlalchemy-declarative-extensions/workflows/test/badge.svg)](https://github.com/dancardin/sqlalchemy-declarative-extensions/actions) [![Coverage Status](https://coveralls.io/repos/github/DanCardin/sqlalchemy-declarative-extensions/badge.svg?branch=main)](https://coveralls.io/github/DanCardin/sqlalchemy-declarative-extensions?branch=main) [![Documentation Status](https://readthedocs.org/projects/sqlalchemy-declarative-extensions/badge/?version=latest)](https://sqlalchemy-declarative-extensions.readthedocs.io/en/latest/?badge=latest)\n\nSee the full documentation [here](https://sqlalchemy-declarative-extensions.readthedocs.io/en/latest/).\n\nAdds extensions to SqlAlchemy (and/or Alembic) which allows declaratively\nstating the existence of additional kinds of objects about your database\nnot natively supported by SqlAlchemy/Alembic.\n\nThis includes:\n\n- Schemas\n- Roles\n- Privileges\n- Rows (i.e. data)\n\nThe primary function(s) of this library include:\n\n- Registering onto the SqlAlchemy event system such that `metadata.create_all`\n  creates these objects.\n- (Optionally) Registers into Alembic such that `alembic revision --autogenerate`\n  automatically creates/updates/deletes declared objects.\n\n## Example Usage\n\n```python\nfrom sqlalchemy import Column, types\nfrom sqlalchemy.orm import as_declarative\nfrom sqlalchemy_declarative_extensions import (\n    declarative_database, Schemas, Roles, Grants, Rows, Row\n)\nfrom sqlalchemy_declarative_extensions.dialects.postgresql import DefaultGrant, Role\n\n\n@declarative_database\n@as_declarative\nclass Base:\n    schemas = Schemas().are("example")\n    roles = Roles(ignore_unspecified=True).are(\n        Role("read", login=False),\n        Role(\n            "app",\n            in_roles=[\'read\']\n        ),\n    )\n    grants = Grants().are(\n        DefaultGrant.on_tables_in_schema("public", \'example\').grant("select", to="read"),\n        DefaultGrant.on_tables_in_schema("public").grant("insert", "update", "delete", to="write"),\n        DefaultGrant.on_sequences_in_schema("public").grant("usage", to="write"),\n    )\n    rows = Rows().are(\n        Row(\'foo\', id=1),\n    )\n\n\nclass Foo(Base):\n    __tablename__ = \'foo\'\n\n    id = Column(types.Integer(), primary_key=True)\n```\n\nNote, there is also support for declaring objects directly through the `MetaData` for\nusers not using sqlalchemy\'s declarative API.\n\n## Event Registration\n\nBy default, the above example does not automatically do anything. Depending on the context,\nyou can use one of two registration hooks: `register_sqlalchemy_events` or `register_alembic_events`.\n\n### `register_sqlalchemy_events`\n\nThis registers events in SqlAlchemy\'s event system such that a `metadata.create_all(...)` call\nwill create the declared database objects.\n\n```python\nfrom sqlalchemy_declarative_extensions import register_sqlalchemy_events\n\nmetadata = Base.metadata  # Given the example above.\nregister_sqlalchemy_events(metadata)\n# Which is equivalent to...\nregister_sqlalchemy_events(metadata, schemas=False, roles=False, grants=False, rows=False)\n```\n\nAll object types are opt in, and should be explicitly included in order to get registered.\n\nPractically, this is to avoid issues with testing. In **most** cases the `metadata.create_all` call\nwill be run in tests, a context where it\'s more likely that you dont **need** grants or grants,\nand where parallel test execution could lead to issues with role or schema creation, depending\non your setup.\n\n### `register_alembic_events`\n\nThis registers comparators in Alembic\'s registration system such that an `alembic revision --autogenerate`\ncommand will diff the existing database state against the declared database objects, and issue\nstatements to create/update/delete objects in order to match the declared state.\n\n```python\n# alembic\'s `env.py`\nfrom sqlalchemy_declarative_extensions import register_alembic_events\n\nregister_alembic_events()\n# Which is equivalent to...\nregister_sqlalchemy_events(schemas=True, roles=True, grants=True, rows=True)\n```\n\nAll object types are opt out but can be excluded.\n\nIn contrast to `register_sqlalchemy_events`, it\'s much more likely that you\'re declaring most of these\nobject types in order to have alembic track them\n\n## Database support\n\nIn principle, this library **can** absolutely support any database supported by SqlAlchemy,\nand capable of being introspected enough to support detection of different kinds of objects.\n\nIn reality, the implementations are going to be purely driven by actual usage. The\ncurrent maintainer(s) primarily use PostgreSQL and as such individual features for\nother databases will either suffer or lack implementation.\n\nAs much as possible, objects will be defined in a database-agnostic way, and the comparison\ninfrastructure should be the sole difference. However databases engines are not the same, and\ncertain kinds of objects, like GRANTs, are inherently database engine specific, and there\'s\nnot much common ground between a PostgreSQL grant and a MySQL one. As such, they will\ninclude database specific objects.\n\n## Alembic-utils\n\nCurrently, the set of supported declarative objects is essentially non-overlapping with\n[Alembic-utils](https://github.com/olirice/alembic_utils). However in principle, there\'s\nno reason that objects supported by this library couldn\'t begin to overlap (views, functions,\ntriggers); and one might begin to question when to use which library.\n\nFirst, it\'s likely that this library can/should grow handlers for objects already supported by\nalembic-utils. In particular, it\'s likely that any future support in this library for something\nlike a view could easily accept an `alembic_utils.pg_view.PGView` definition and handle it directly.\nThe two libraries are likely fairly complementary in that way, although it\'s important to note\nsome of the differences.\n\nAlembic utils:\n\n- Is more directly tied to Alembic and specifically provides functionality for autogenerating\n  DDL for alembic, as the name might imply. It does **not** register into sqlalchemy\'s event\n  system.\n- Requires one to explicitly find/include the objects one wants to track with alembic.\n- It provides direct translation of individual entities (like a single, specific `PGGrantTable`).\n- In most cases, it appears to define a very "literal" interface (for example, `PGView` accepts\n  the whole view definition as a raw literal string), rather than an abstracted one.\n\nBy contrast, this library:\n\n- SqlAlchemy is the main dependency and registration point. The primary function of the library\n  is to register into sqlalchemy\'s event system to ensure that a `metadata.create_all` performs\n  the requisite statements to ensure the state of the database matches the declaration.\n\n  This library does **not** require alembic, but it does (optionally) perform a similar function\n  by way of enabling autogeneration support for non-native objects.\n\n- Perhaps a technical detail, but this library registers the declaratively stated objects directly\n  on the metadata/declarative-base. This allows the library to automatically know the intended\n  state of the world, rather than needing to discover objects.\n- The intended purpose of the supported objects is to declare what the state of the world **should**\n  look like. Therefore the function of this library includes the (optional) **removal** of objects\n  detected to exist which are not declared (much like alembic does for tables). Whereas alembic-utils\n  only operates on objects you create entities for.\n- As much as possible, this library provides more abstracted interfaces for defining objects.\n  This is particularly important for objects like roles/grants where not every operation is a create\n  or delete (in contrast to something like a view).\n',
    'author': 'Dan Cardin',
    'author_email': 'ddcardin@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dancardin/sqlalchemy-declarative-extensions',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



from typing import List, Optional, Tuple, TypeVar

from sqlalchemy_declarative_extensions.dialects.postgresql import (
    DefaultGrant,
    DefaultGrantStatement,
    DefaultGrantTypes,
    Grant,
    GrantStatement,
    GrantTypes,
)

GT = TypeVar("GT", DefaultGrantTypes, GrantTypes)


def parse_acl(
    acl: Optional[str],
    object_type: str,
    object_name: str,
    owner: Optional[str] = None,
    expanded: bool = False,
) -> List[GrantStatement]:
    type = GrantTypes.from_relkind(object_type)
    _, grants = _parse_acl(acl, type, owner=owner, expanded=expanded)

    return [
        GrantStatement(
            grant=grant,
            grant_type=type,
            targets=(object_name,),
        )
        for grant in grants
    ]


def parse_default_acl(
    acl: Optional[str],
    object_type: str,
    schema_name: str = "public",
    expanded: bool = False,
    current_role: Optional[str] = None,
) -> List[DefaultGrantStatement]:
    type = DefaultGrantTypes.from_relkind(object_type)

    grantor: Optional[str]
    grantor, grants = _parse_acl(acl, type, expanded=expanded)

    if current_role == grantor:
        grantor = None

    return [
        DefaultGrantStatement(
            grant=grant,
            default_grant=DefaultGrant(
                grant_type=type,
                in_schemas=(schema_name,),
                target_role=grantor,
            ),
        )
        for grant in grants
    ]


def _parse_acl(
    acl: Optional[str],
    type: GT,
    *,
    owner: Optional[str] = None,
    expanded: bool = False,
) -> Tuple[str, List[Grant]]:
    """Parse postgres' acl mini language.

    Port of ``parseAclItem`` from `dumputils.c`.
    """
    variants = type.to_variants()

    grants_no_grant_option: List[Grant] = []
    grants_with_grant_option: List[Grant] = []

    if acl is not None:
        eq_pos, grantee = get_acl_username(acl)
        assert acl[eq_pos] == "="

        if not grantee:
            grantee = "PUBLIC"

        slash_pos = acl.index("/", eq_pos)
        _, grantor = get_acl_username(acl[slash_pos + 1 :])

        priv = acl[eq_pos + 1 : slash_pos]

        for privilege, symbol in variants.acl_symbols().items():
            pos = priv.find(symbol)

            if pos >= 0:
                if len(priv) > pos + 1 and priv[pos + 1] == "*":
                    grants_with_grant_option.append(privilege)
                else:
                    grants_no_grant_option.append(privilege)
    else:
        assert owner is not None
        priv = None
        grantor = owner
        grantee = owner
        grants_no_grant_option.extend(list(variants))

    grants = []
    for grant_option, grant_privileges in (
        (False, grants_no_grant_option),
        (True, grants_with_grant_option),
    ):
        if not grant_privileges:
            continue

        # By putting the grants in a group by itself before iterating over it,
        # non-`expanded` versions will take the whole set as a single group.
        if expanded:
            grant_groups = [[g] for g in grant_privileges]
        else:
            grant_groups = [grant_privileges]

        for grant_group in grant_groups:
            grant = Grant(
                target_role=grantee,
                grants=tuple(sorted(grant_group)),  # type: ignore
                grant_option=grant_option,
            )
            grants.append(grant)

    if priv:
        assert len(grants) > 0

    return grantor, grants


def get_acl_username(acl: str):
    """Port ``copyAclUserName`` from ``dumputils.c``."""
    i = 0
    output = ""

    acl_len = len(acl)
    while i < acl_len and acl[i] != "=":
        # If user name isn't quoted, then just add it to the output buffer
        if acl[i] != '"':
            output += acl[i]
            i += 1
        else:
            i += 1

            # Loop until we come across an unescaped quote
            while True:
                if i >= acl_len:
                    raise ValueError("ACL syntax error: unterminated quote.")

                char = acl[i]

                if char == '"':
                    # Quoting convention is to escape " as "".
                    has_next_char = i + 1 < acl_len
                    if not has_next_char:
                        break

                    next_char = acl[i + 1]
                    i += 1
                    if next_char != '"':
                        break

                output += char
                i += 1

    return i, output



from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import Generator, Iterable, List, Optional, Union

from sqlalchemy_declarative_extensions.dialects import postgresql
from sqlalchemy_declarative_extensions.role import generic


@dataclass
class Roles:
    roles: List[Union[postgresql.Role, generic.Role]] = field(default_factory=list)
    ignore_unspecified: bool = False
    ignore_roles: List[str] = field(default_factory=list)

    @classmethod
    def coerce_from_unknown(
        cls,
        unknown: Union[
            None, Iterable[Union[postgresql.Role, generic.Role, str]], Roles
        ],
    ) -> Optional[Roles]:
        if isinstance(unknown, Roles):
            return unknown

        if isinstance(unknown, Iterable):
            return Roles().are(*unknown)

        return None

    def __iter__(self) -> Generator[Union[postgresql.Role, generic.Role], None, None]:
        for role in self.roles:
            yield role

    def are(self, *roles: Union[postgresql.Role, generic.Role, str]):
        return replace(
            self,
            roles=[generic.Role.coerce_from_unknown(role) for role in roles],
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass(frozen=True)
class Role:
    name: str

    in_roles: Optional[List[str]] = None

    @classmethod
    def coerce_from_unknown(cls, unknown: Union[str, Role]) -> Role:
        if isinstance(unknown, Role):
            return unknown

        return cls(unknown)

    @property
    def has_option(self):
        return False

    @property
    def options(self):
        return []

    def to_sql_create(self) -> str:
        statement = f'CREATE ROLE "{self.name}"'
        if self.in_roles is not None:
            in_roles = ", ".join(self.in_roles)
            statement += f"IN ROLE {in_roles}"
        return statement + ";"

    def to_sql_update(self, to_role) -> List[str]:
        raise NotImplementedError(
            "When using the generic role, there should never exist any cause to update a role."
        )

    def to_sql_drop(self) -> str:
        return f'DROP ROLE "{self.name}";'



from sqlalchemy.engine import Connection


def dialect_dispatch(postgresql=None, sqlite=None):
    dispatchers = {
        "postgresql": postgresql,
        "sqlite": sqlite,
    }

    def dispatch(connection: Connection, **kwargs):
        dialect_name = connection.dialect.name
        if dialect_name == "pmrsqlite":
            dialect_name = "sqlite"

        dispatcher = dispatchers.get(dialect_name)
        if dispatcher is None:  # pragma: no cover
            raise NotImplementedError(
                f"'{dialect_name}' is not yet supported for this operation."
            )

        return dispatcher(connection, **kwargs)

    return dispatch



from __future__ import annotations

from dataclasses import dataclass, field, fields
from datetime import datetime
from typing import List, Optional, Union

from sqlalchemy_declarative_extensions.role import generic


@dataclass(frozen=True)
class Role(generic.Role):
    """Define a role object.

    postgres: https://www.postgresql.org/docs/current/sql-createrole.html

    Note, the password field can be supplied, but it will be ignored when
    performing comparisons against existing roles. That is, changing the
    password field will not produce any (alembic) changes!
    """

    superuser: Optional[bool] = False
    createdb: Optional[bool] = False
    createrole: Optional[bool] = False
    inherit: Optional[bool] = True
    login: Optional[bool] = False
    replication: Optional[bool] = False
    bypass_rls: Optional[bool] = False
    connection_limit: Optional[int] = None
    valid_until: Optional[datetime] = None

    password: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_pg_role(cls, r) -> Role:
        return cls(
            r.rolname,
            superuser=r.rolsuper,
            createdb=r.rolcreatedb,
            createrole=r.rolcreaterole,
            login=r.rolcanlogin,
            replication=r.rolreplication,
            connection_limit=r.rolconnlimit if r.rolconnlimit != -1 else None,
            bypass_rls=r.rolbypassrls,
            valid_until=r.rolvaliduntil,
            in_roles=r.memberof or None,
        )

    @classmethod
    def from_unknown_role(cls, r: Union[generic.Role, Role]) -> Role:
        if not isinstance(r, Role):
            return Role(r.name, in_roles=r.in_roles)

        return r

    @property
    def options(self):
        for f in fields(self):
            if f.name == "name":
                continue

            value = getattr(self, f.name)
            if value is None:
                continue

            yield f.name, value

    def __repr__(self):
        cls_name = self.__class__.__name__
        options = ", ".join([f"{key}={repr(value)}" for key, value in self.options])
        return f'{cls_name}("{self.name}", {options})'

    def to_sql_create(self) -> str:
        segments = ["CREATE ROLE", self.name]

        options = postgres_render_role_options(self)
        if options:
            segments.append("WITH")
        segments.extend(options)

        if self.in_roles is not None:
            in_roles = ", ".join(self.in_roles)
            segment = f"IN ROLE {in_roles}"
            segments.append(segment)

        command = " ".join(segments)
        return command + ";"

    def to_sql_update(self, to_role: Role) -> List[str]:
        role_name = to_role.name
        diff = RoleDiff.diff(self, to_role)
        segments = ["ALTER ROLE", role_name, "WITH"]
        segments.extend(postgres_render_role_options(diff))

        alter = " ".join(segments) + ";"
        result = [alter]
        for add_name in diff.add_roles:
            result.append(f"GRANT {add_name} to {role_name};")

        for remove_name in diff.remove_roles:
            result.append(f"GRANT {remove_name} to {role_name};")

        return result


@dataclass
class RoleDiff:
    name: str
    superuser: Optional[bool] = None
    createdb: Optional[bool] = None
    createrole: Optional[bool] = None
    inherit: Optional[bool] = None
    login: Optional[bool] = None
    replication: Optional[bool] = None
    bypass_rls: Optional[bool] = None
    connection_limit: Optional[int] = None
    valid_until: Optional[datetime] = None

    add_roles: List[str] = field(default_factory=list)
    remove_roles: List[str] = field(default_factory=list)

    @classmethod
    def diff(cls, from_role: Role, to_role: Role) -> RoleDiff:
        superuser = None
        if to_role.superuser != from_role.superuser:
            superuser = to_role.superuser

        createdb = None
        if to_role.createdb != from_role.createdb:
            createdb = to_role.createdb

        createrole = None
        if to_role.createrole != from_role.createrole:
            createrole = to_role.createrole

        inherit = None
        if to_role.inherit != from_role.inherit:
            inherit = to_role.inherit

        login = None
        if to_role.login != from_role.login:
            login = to_role.login

        replication = None
        if to_role.replication != from_role.replication:
            replication = to_role.replication

        bypass_rls = None
        if to_role.bypass_rls != from_role.bypass_rls:
            bypass_rls = to_role.bypass_rls

        connection_limit = None
        if to_role.connection_limit != from_role.connection_limit:
            connection_limit = to_role.connection_limit

        valid_until = None
        if to_role.valid_until != from_role.valid_until:
            valid_until = to_role.valid_until

        add_roles = []
        remove_roles = []
        if to_role.in_roles != from_role.in_roles:
            from_in_roles = from_role.in_roles or []
            to_in_roles = to_role.in_roles or []

            add_roles = [r for r in to_in_roles if r not in from_in_roles]
            remove_roles = [r for r in from_in_roles if r not in to_in_roles]

        return cls(
            name=to_role.name,
            superuser=superuser,
            createdb=createdb,
            createrole=createrole,
            inherit=inherit,
            login=login,
            replication=replication,
            bypass_rls=bypass_rls,
            connection_limit=connection_limit,
            valid_until=valid_until,
            add_roles=add_roles,
            remove_roles=remove_roles,
        )


def conditional_option(option, condition):
    if not condition:
        option = f"NO{option}"
    return option


def postgres_render_role_options(role: Union[Role, RoleDiff]) -> List[str]:
    segments = []

    if role.superuser is not None:
        segment = conditional_option("SUPERUSER", role.superuser)
        segments.append(segment)

    if role.createdb is not None:
        segment = conditional_option("CREATEDB", role.createdb)
        segments.append(segment)

    if role.createrole is not None:
        segment = conditional_option("CREATEROLE", role.createrole)
        segments.append(segment)

    if role.inherit is not None:
        segment = conditional_option("INHERIT", role.inherit)
        segments.append(segment)

    if role.login is not None:
        segment = conditional_option("LOGIN", role.login)
        segments.append(segment)

    if role.replication is not None:
        segment = conditional_option("REPLICATION", role.replication)
        segments.append(segment)

    if role.bypass_rls is not None:
        segment = conditional_option("BYPASSRLS", role.bypass_rls)
        segments.append(segment)

    if role.connection_limit is not None:
        segment = f"CONNECTION LIMIT {role.connection_limit}"
        segments.append(segment)

    if isinstance(role, Role) and role.password is not None:
        segment = f"PASSWORD {role.password}"
        segments.append(segment)

    if role.valid_until is not None:
        timestamp = role.valid_until.isoformat()
        segment = f"VALID UNTIL '{timestamp}'"
        segments.append(segment)

    return segments
